/**
   Sample picking up motion controller for PA10.
   @author Shin'ichiro Nakaoka
*/

#include "Interpolator.h"
#include <cnoid/SimpleController>
#include <cnoid/Link>
#include <cnoid/JointPath>
#include <cnoid/EigenUtil>
#include <cmath>

using namespace std;
using namespace cnoid;
using namespace boost;

static double pgain[] = {
    35000.0, 35000.0, 35000.0, 35000.0, 35000.0, 35000.0, 35000.0,
    17000.0, 17000.0 };

static double dgain[] = {
    400.0, 400.0, 400.0, 400.0, 400.0, 400.0, 400.0,
    400.0, 400.0 };

typedef Eigen::Matrix<double, 6, 1> Vector6;

class PA10PickupController : public cnoid::SimpleController
{
    int phase;
    //Interpolator<Vector6> wristInterpolator; // cause alignment problem (seg fault)
    Interpolator<VectorXd> wristInterpolator;
    Interpolator<VectorXd> jointInterpolator;
    Link* wrist;
    Link* leftHand;
    Link* rightHand;
    JointPathPtr baseToWrist;
    double dq_hand;
    VectorXd qref, qold, qref_old;
    double time;
    double timeStep;

public:

    Vector3 toRadianVector3(double x, double y, double z){
        return Vector3(radian(x), radian(y), radian(z));
    }
    
    virtual bool initialize(Body* body, double timeStep, ostream& os) {

        this->timeStep = timeStep;
        time = 0.0;

        int n = body->numJoints();
        qref.resize(n);
        qref_old.resize(n);
        qold.resize(n);

        for(int i=0; i < n; ++i){
            qold[i] = body->joint(i)->q;
        }
        qref = qold;
        qref_old = qold;

        wrist = body->link("J7");
        baseToWrist = body->getJointPath(body->rootLink(), wrist);

        leftHand  = body->link("HAND_L");
        rightHand = body->link("HAND_R");

        //Vector6 p0;
        VectorXd p0(6);
        p0.head<3>() = wrist->p;
        p0.tail<3>() = rpyFromRot(wrist->attitude());

        //Vector6 p1;
        VectorXd p1(6);
        p1.head<3>() = Vector3(0.9, 0.0, 0.25);
        p1.tail<3>() = toRadianVector3(180.0, 0.0, 0.0);

        wristInterpolator.clear();
        wristInterpolator.appendSample(0.0, p0);
        wristInterpolator.appendSample(1.0, p1);
        p1.z() = 0.2;
        wristInterpolator.appendSample(1.2, p1);
        wristInterpolator.update();

        phase = 0;
        dq_hand = 0.0;

        return true;
    }

    virtual bool control(Body* body, ostream& os) {

        VectorXd q(body->numJoints());
        for(int i=0; i < body->numJoints(); ++i){
            q[i] = body->joint(i)->q;
            body->joint(i)->q = qref_old[i];
        }

        //Vector6 p;
        VectorXd p(6);

        if(phase <= 3){
            p  = wristInterpolator.interpolate(time);
            if(baseToWrist->calcInverseKinematics(
                   Vector3(p.head<3>()), wrist->calcRfromAttitude(rotFromRpy(Vector3(p.tail<3>()))))){
                for(int i=0; i < baseToWrist->numJoints(); ++i){
                    Link* joint = baseToWrist->joint(i);
                    qref[joint->jointId] = joint->q;
                }
            }
        }

        if(phase == 0){
            if(time > wristInterpolator.domainUpper()){
                phase = 1;
            }

        } else if(phase == 1){
            if(fabs(rightHand->u) < 40.0 || fabs(leftHand->u) < 40.0){ // not holded ?
                dq_hand = std::min(dq_hand + 0.00001, 0.0005);
                qref[rightHand->jointId] -= radian(dq_hand);
                qref[leftHand->jointId]  += radian(dq_hand);

            } else {
                //Vector6 p2;
                VectorXd p2(6);
                p2.head<3>() = Vector3(0.0, 0.5, 1.0);
                p2.tail<3>() = toRadianVector3(180.0, -45, 90.0);

                //Vector6 p3;
                VectorXd p3(6);
                p3.head<3>() = Vector3(0.0, 0.7, 0.52);
                p3.tail<3>() = toRadianVector3(180.0, 0, 90.0);

                wristInterpolator.clear();
                wristInterpolator.appendSample(time, p);
                wristInterpolator.appendSample(time + 1.0, p2);
                wristInterpolator.appendSample(time + 1.5, p3);
                wristInterpolator.appendSample(time + 1.7, p3);
                wristInterpolator.update();
                phase = 2;
            }
        } else if(phase == 2){
            if(time > wristInterpolator.domainUpper()){
                phase = 3;
                dq_hand = 0.0;
            }
        } else if(phase == 3){
            if(qref[rightHand->jointId] < 0.028 || qref[leftHand->jointId] > -0.028){
                dq_hand = std::min(dq_hand + 0.00001, 0.002);
                qref[rightHand->jointId] += radian(dq_hand);
                qref[leftHand->jointId]  -= radian(dq_hand);
            } else {
                jointInterpolator.clear();
                jointInterpolator.appendSample(time, qref);
                VectorXd qf = VectorXd::Zero(qref.size());
                qf[rightHand->jointId] = qref[rightHand->jointId];
                qf[leftHand->jointId]  = qref[leftHand->jointId];
                jointInterpolator.appendSample(time + 1.0, qf);
                jointInterpolator.update();
                phase = 4;
            }
        } else if(phase == 4){
            qref = jointInterpolator.interpolate(time);
        }
            
        for(int i=0; i < body->numJoints(); ++i){
            double dq = (q[i] - qold[i]) / timeStep;
            double dq_ref = (qref[i] - qref_old[i]) / timeStep;
            body->joint(i)->u = (qref[i] - q[i]) * pgain[i] + (dq_ref - dq) * dgain[i];
        }
        qold = q;
        qref_old = qref;
        time += timeStep;

        return true;
    }
};


CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(PA10PickupController)
